
from typing import cast

from logging import Logger
from logging import getLogger

from dataclasses import dataclass

from wx import CANCEL
from wx import EVT_BUTTON
from wx import EVT_CLOSE
from wx import ID_CANCEL
from wx import ID_OK
from wx import OK
from wx import DEFAULT_DIALOG_STYLE
from wx import STAY_ON_TOP

from wx import CommandEvent
from wx import Size
from wx import Window
from wx import StdDialogButtonSizer

from wx.lib.sized_controls import SizedDialog
from wx.lib.sized_controls import SizedPanel

from umlextensions.ExtensionsPreferences import ExtensionsPreferences
from umlextensions.input.python.NamedSlider import NamedSlider


@dataclass
class ShapeLayout:
    startX:     int = 0
    startY:     int = 0
    xIncrement: int = 0
    maximumX:   int = 0

class DlgShapeLayoutParameters(SizedDialog):
    def __init__(self, parent: Window = None):

        style: int = STAY_ON_TOP | DEFAULT_DIALOG_STYLE

        super().__init__(parent=parent, title='Layout Parameters', style=style, size=Size(width=300, height=600))

        self.logger: Logger = getLogger(__name__)

        self._extensionPrefs: ExtensionsPreferences = ExtensionsPreferences()

        sizedPanel: SizedPanel = self.GetContentsPane()
        sizedPanel.SetSizerType('vertical')
        sizedPanel.SetSizerProps(expand=True, proportion=1)

        self._startX:     NamedSlider = cast(NamedSlider, None)
        self._startY:     NamedSlider = cast(NamedSlider, None)
        self._xIncrement: NamedSlider = cast(NamedSlider, None)
        self._maximumX:   NamedSlider = cast(NamedSlider, None)

        self._shapeLayout: ShapeLayout = ShapeLayout(
            startX=self._extensionPrefs.startX,
            startY=self._extensionPrefs.startY,
            xIncrement=self._extensionPrefs.xIncrement,
            maximumX=self._extensionPrefs.maximumX
        )
        self._layoutControls(parentPanel=sizedPanel)
        self._layoutStandardOkCancelButtonSizer()

        self._setControlValues()
        # CallAfter(self._resizeDialog)
        # self._resizeDialog()

    @property
    def shapeLayout(self) -> ShapeLayout:
        return self._shapeLayout

    def _layoutControls(self, parentPanel: SizedPanel):
        """
        Args:
            parentPanel:
        """
        self._startX     = NamedSlider(sizedPanel=parentPanel, label='Shape Start X',     minValue=20,   maxValue=500)
        self._startY     = NamedSlider(sizedPanel=parentPanel, label='Shape Start Y',     minValue=20,   maxValue=500)
        self._xIncrement = NamedSlider(sizedPanel=parentPanel, label='Shape X Increment', minValue=20,   maxValue=500)
        self._maximumX   = NamedSlider(sizedPanel=parentPanel, label='Maximum X',         minValue=3000, maxValue=20000)

    def _setControlValues(self):

        self._startX.value     = self._shapeLayout.startX
        self._startY.value     = self._shapeLayout.startY
        self._xIncrement.value = self._shapeLayout.xIncrement
        self._maximumX.value   = self._shapeLayout.maximumX

    def _layoutStandardOkCancelButtonSizer(self):
        """
        Call this last when creating controls;  Will take care of
        adding callbacks for the Ok and Cancel buttons
        """
        buttSizer: StdDialogButtonSizer = self.CreateStdDialogButtonSizer(OK | CANCEL)

        self.SetButtonSizer(buttSizer)
        self.Bind(EVT_BUTTON, self._onOk,    id=ID_OK)
        self.Bind(EVT_BUTTON, self._onClose, id=ID_CANCEL)
        self.Bind(EVT_CLOSE,  self._onClose)

    # noinspection PyUnusedLocal
    def _onOk(self, event: CommandEvent):
        """
        Save the values in the component return value and persist them in the preferences
        Args:
            event:
        """

        self._extensionPrefs.startX     = self._startX.value
        self._extensionPrefs.startY     = self._startY.value
        self._extensionPrefs.xIncrement = self._xIncrement.value
        self._extensionPrefs.maximumX   = self._maximumX.value

        self._shapeLayout.startX     = self._startX.value
        self._shapeLayout.startY     = self._startY.value
        self._shapeLayout.xIncrement = self._xIncrement.value
        self._shapeLayout.maximumX   = self._maximumX.value

        self.EndModal(OK)

    # noinspection PyUnusedLocal
    def _onClose(self, event: CommandEvent):
        self.EndModal(CANCEL)

    def _resizeDialog(self):
        """
        A little trick to make sure that you can't resize the dialog to
        less screen space than the controls need
        """
        self.Fit()
        self.SetMinSize(self.GetSize())




from logging import Logger
from logging import getLogger

from umlextensions.IExtensionsFacade import IExtensionsFacade
from umlextensions.extensiontypes.BaseExtension import BaseExtension


class BaseOutputExtension(BaseExtension):
    """
        Base class for extensions that can convert UML Diagrams into foreign
        structured data.  Examples include but are not limited to:

        * Mermaid
        * PDF
        * Images (png, jpg, bmp, etc
        * Generate Python code
        * Generate Java code

    """
    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade)
        self.logger: Logger = getLogger(__name__)




from typing import List
from typing import Dict
from typing import Union
from typing import NewType

from enum import Enum

from dataclasses import dataclass

from umlmodel.Class import Class

VERSION: str = '3.0'

ModelClassName = NewType('ModelClassName', str)
ParentName     = NewType('ParentName',     str)
PropertyName   = NewType('PropertyName',   str)
ChildName      = NewType('ChildName',      str)

PropertyNames = NewType('PropertyNames', List[PropertyName])
ModelClasses  = NewType('ModelClasses',  Dict[ModelClassName, Class])

PropertyMap   = NewType('PropertyMap', Dict[ModelClassName, PropertyNames])
Children      = List[Union[ModelClassName, ChildName]]
Parents       = NewType('Parents',        Dict[ParentName,    Children])

AssociateName = ModelClassName

class AssociationType(Enum):

    ASSOCIATION = 'ASSOCIATION'
    AGGREGATION = 'AGGREGATION'
    COMPOSITION = 'COMPOSITION'
    INHERITANCE = 'INHERITANCE'
    INTERFACE   = 'INTERFACE'

@dataclass
class Associate:
    associateName:   AssociateName   = AssociateName('')
    associationType: AssociationType = AssociationType.ASSOCIATION


Associates = NewType('Associates', List[Associate])

#
# e.g.
#     @property
#     def pages(self) -> Pages:
#         return self._pages
# In the above "Pages" is the AssociateName and goes in the List for the method containing ModelClassName
#
# e.g.
#  self.pages: Pages = Pages({})
#
#  Pages is the AssociateName and the enclosing class for the __init__ method is the ModelClassName
#
#
Associations = NewType('Associations', Dict[ModelClassName, Associates])




from sys import exc_info

from traceback import format_exception

MAX_STACK: int = 5


class ErrorFormatter:
    """
    A static class with class methods to simplify error reporting
    """
    def __init__(self):
        pass

    @classmethod
    def getErrorMessage(cls):
        error, eMessage, eTraceback = exc_info()
        return str(eMessage)

    @classmethod
    def getErrorStack(cls, e, limit: int = MAX_STACK) -> str:
        """
        Returns the bottom half of the error stack
        Args:
            e:
            limit:

        Returns:

        """

        errorList = format_exception(e)
        errorList.reverse()

        fs:         str = ''
        stackLimit: int = 0

        for e in errorList:
            fs = f'{fs}{e}'
            stackLimit += 1
            if stackLimit == limit:
                break

        return fs




from typing import List
from typing import cast

from logging import Logger
from logging import getLogger

from wx import OK
from wx import ICON_ERROR
from wx import PD_APP_MODAL
from wx import PD_ELAPSED_TIME

from wx import BeginBusyCursor
from wx import EndBusyCursor
from wx import ProgressDialog
from wx import MessageBox
from wx import MessageDialog

from wx import Yield as wxYield

from umlshapes.ShapeTypes import UmlLinks

from umlshapes.shapes.UmlClass import UmlClass

from umlshapes.types.UmlDimensions import UmlDimensions
from umlshapes.types.UmlPosition import UmlPosition

from umlshapes.frames.ClassDiagramFrame import ClassDiagramFrame

from umlextensions.IExtensionsFacade import IExtensionsFacade

from umlextensions.extensiontypes.ExtensionDataTypes import Author
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionDescription
from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionName
from umlextensions.extensiontypes.ExtensionDataTypes import FileSuffix
from umlextensions.extensiontypes.ExtensionDataTypes import FormatName
from umlextensions.extensiontypes.ExtensionDataTypes import Version

from umlextensions.input.InputFormat import InputFormat

from umlextensions.input.BaseInputExtension import BaseInputExtension

from umlextensions.input.python.DlgSelectMultiplePackages import DlgSelectMultiplePackages
from umlextensions.input.python.DlgSelectMultiplePackages import ImportPackages
from umlextensions.input.python.DlgSelectMultiplePackages import Package
from umlextensions.input.python.DlgShapeLayoutParameters import DlgShapeLayoutParameters
from umlextensions.input.python.DlgShapeLayoutParameters import ShapeLayout

from umlextensions.input.python.PythonParseException import PythonParseException
from umlextensions.input.python.PythonToUmlShapes import PythonToUmlShapes
from umlextensions.input.python.PythonToUmlShapes import UmlClassesDict

from umlextensions.input.python.visitor.ParserTypes import ModelClasses

FORMAT_NAME:           FormatName           = FormatName("Python File(s)")
FILE_SUFFIX:           FileSuffix           = FileSuffix('py')
EXTENSION_DESCRIPTION: ExtensionDescription = ExtensionDescription('Python code reverse engineering')

class InputPython(BaseInputExtension):

    def __init__(self, extensionsFacade: IExtensionsFacade):

        super().__init__(extensionsFacade)
        self.logger: Logger = getLogger(__name__)

        self._name    = ExtensionName('Python to UML')
        self._author  = Author('Humberto A. Sanchez II')
        self._version = Version('3.0')

        self._inputFormat  = InputFormat(formatName=FORMAT_NAME, fileSuffix=FILE_SUFFIX, description=EXTENSION_DESCRIPTION)

        self._packageCount:   int = 0
        self._moduleCount:    int = 0
        self._importPackages: ImportPackages = ImportPackages([])

        self._readProgressDlg: ProgressDialog = cast(ProgressDialog, None)

    def setImportOptions(self) -> bool:
        """
        We do need to ask for the input file names

        TODO:  super complicated if else logic

        Returns:  'True', we support import
        """
        if isinstance(self._frameInformation.umlFrame, ClassDiagramFrame) is True:
            startDirectory: str = self._preferences.startDirectory
            with DlgSelectMultiplePackages(startDirectory=startDirectory, inputFormat=self.inputFormat) as dlg:
                if dlg.ShowModal() == OK:
                    self._packageCount   = dlg.packageCount
                    self._moduleCount    = dlg.moduleCount
                    self._importPackages = dlg.importPackages
                    #
                    # Ensure we picked at least 1
                    if self._packageCount == 0:
                        return False
                    else:
                        return True
                else:
                    return False
        else:
            booBoo: MessageDialog = MessageDialog(parent=None,
                                                  message='The import to frame must be a class diagram frame',
                                                  caption='Error!',
                                                  style=OK | ICON_ERROR
                                                  )
            booBoo.ShowModal()
            return False

    def read(self) -> bool:

        BeginBusyCursor()
        wxYield()
        status: bool = True
        try:
            self._readProgressDlg = ProgressDialog('Parsing Files', 'Starting', parent=None, style=PD_APP_MODAL | PD_ELAPSED_TIME)

            # reverseEngineer: ReverseEngineerPythonV3 = ReverseEngineerPythonV3()
            # Should the extensions know about the UML pub/sub Engine ???
            #
            self._readProgressDlg.SetRange(self._moduleCount)

            classDiagramFrame: ClassDiagramFrame = cast(ClassDiagramFrame, self._frameInformation.umlFrame)
            pythonToUmlShapes: PythonToUmlShapes = PythonToUmlShapes(classDiagramFrame=classDiagramFrame, umlPubSubEngine=self._extensionsFacade.umlPubSub)

            modelClasses: ModelClasses = self._collectModelClassesInPass1(pythonToUmlShapes=pythonToUmlShapes)
            modelClasses              = self._enhanceModelClassesInPass2(pythonToUmlShapes=pythonToUmlShapes, modelClasses=modelClasses)

            umlClassesDict: UmlClassesDict = pythonToUmlShapes.generateUmlClasses(modelClasses)
            pythonToUmlShapes.generateLinks(umlClassesDict)

            self._readProgressDlg.Destroy()
            self._layoutUmlClasses(umlClasses=list(umlClassesDict.values()))
            self._layoutLinks(oglLinks=pythonToUmlShapes.umlLinks)

        except (ValueError, Exception, PythonParseException) as e:
            self._readProgressDlg.Destroy()
            MessageBox(f'{e}', 'Error', OK | ICON_ERROR)
            status = False
        else:
            self._extensionsFacade.extensionModifiedProject()
        finally:
            EndBusyCursor()
            self._extensionsFacade.refreshFrame()
            wxYield()

        self._extensionsFacade.wiggleShapes()

        return status

    def _collectModelClassesInPass1(self, pythonToUmlShapes: PythonToUmlShapes) -> ModelClasses:

        cumulativeModelClasses: ModelClasses = ModelClasses({})
        for directory in self._importPackages:
            importPackage: Package = cast(Package, directory)

            currentModelClasses: ModelClasses = pythonToUmlShapes.pass1(directoryName=importPackage.packageName,
                                                                        files=importPackage.importModules,
                                                                        progressCallback=self._readProgressCallback)

            cumulativeModelClasses = ModelClasses(cumulativeModelClasses | currentModelClasses)

        return cumulativeModelClasses

    def _enhanceModelClassesInPass2(self, pythonToUmlShapes: PythonToUmlShapes, modelClasses: ModelClasses) -> ModelClasses:

        updatedModelClasses: ModelClasses = ModelClasses({})
        for directory in self._importPackages:
            importPackage: Package = cast(Package, directory)

            updatedModelClasses = pythonToUmlShapes.pass2(directoryName=importPackage.packageName,
                                                          files=importPackage.importModules,
                                                          modelClasses=modelClasses,
                                                          progressCallback=self._readProgressCallback
                                                          )

        return updatedModelClasses

    def _readProgressCallback(self, currentFileCount: int, msg: str):
        """

        Args:
            currentFileCount:   The current file # we are working pm
            msg:    An updated message
        """

        self._readProgressDlg.Update(currentFileCount, msg)

    def _layoutUmlClasses(self, umlClasses: List[UmlClass]):
        """
        Organize by vertical descending sizes

        Args:
            umlClasses
        """
        # Sort by descending height
        # noinspection PyProtectedMember
        sortedUmlClasses = sorted(umlClasses, key=lambda umlClassToSort: umlClassToSort.GetHeight(), reverse=True)

        with DlgShapeLayoutParameters() as dlg:
            if dlg.ShowModal() == OK:
                self.logger.info('Ok')

            # x: int = 20     # startX
            # y: int = 20     # startY
            shapeLayout: ShapeLayout = dlg.shapeLayout
            x: int = shapeLayout.startX
            y: int = shapeLayout.startY

            incY: int = 0
            for umlClass in sortedUmlClasses:
                #  incX, sy = oglClass.GetSize()
                size: UmlDimensions = umlClass.size
                incX = size.width
                sy   = size.height
                incX += shapeLayout.xIncrement         # xIncrement
                sy += 20
                incY = max(incY, int(sy))              # find good coordinates
                if x + incX >= shapeLayout.maximumX:   # maximumX
                    x = 20
                    y += incY
                    incY = int(sy)

                # oglClass.SetPosition(x, y)
                umlClass.position = UmlPosition(x=x, y=y)
                x += incX
                # self._pluginAdapter.addShape(shape=oglClass)
                self._extensionsFacade.addShape(umlShape=umlClass)

    def _layoutLinks(self, oglLinks: UmlLinks):
        for oglLink in oglLinks:
            self._extensionsFacade.addShape(oglLink)




from dataclasses import dataclass

from umlextensions.extensiontypes.BaseRequestResponse import BaseRequestResponse


@dataclass
class ImportDirectoryResponse(BaseRequestResponse):
    directoryName: str = ''




from logging import Logger
from logging import getLogger

from wx import ID_ANY
from wx import SL_AUTOTICKS
from wx import SL_HORIZONTAL
from wx import SL_LABELS
from wx import VERTICAL

from wx import Panel
from wx import Size
from wx import SizerFlags
from wx import Slider
from wx import BoxSizer

from wx.lib.sized_controls import SizedPanel
from wx.lib.sized_controls import SizedStaticBox

SLIDER_HEIGHT: int = 100
SLIDER_WIDTH:  int = 300


class NamedSlider(Slider):

    def __init__(self, sizedPanel: SizedPanel, label: str, minValue: int, maxValue: int):
        """
        The reason with have a work around panel
        https://discuss.wxpython.org/t/wxpython-slider-incorrectly-displays-with-sized-panels/36915/10

        Args:
            sizedPanel:
            label:
            minValue:
            maxValue:
        """
        self.logger: Logger = getLogger(__name__)

        namingPanel: SizedStaticBox = SizedStaticBox(sizedPanel, label=label)
        namingPanel.SetSizerType('horizontal')
        namingPanel.SetSizerProps(expand=True, proportion=1)

        workAroundPanel: Panel    = Panel(namingPanel)
        workAroundSizer: BoxSizer = BoxSizer(VERTICAL)
        workAroundPanel.SetSizer(workAroundSizer)

        super().__init__(
            parent=workAroundPanel, id=ID_ANY,
            minValue=minValue, maxValue=maxValue,
            size=Size(width=SLIDER_WIDTH, height=SLIDER_HEIGHT),
            style=SL_HORIZONTAL | SL_AUTOTICKS | SL_LABELS
        )
        flagsExpand: SizerFlags = SizerFlags().Expand().Proportion(1)
        workAroundSizer.Add(window=self, flags=flagsExpand)

    @property
    def value(self) -> int:
        return self.GetValue()

    @value.setter
    def value(self, value: int):
        self.SetValue(value)




from typing import cast
from typing import List
from typing import NewType

from logging import Logger
from logging import getLogger

from re import search as regExSearch
from re import Match as regExMatch

from dataclasses import dataclass

from antlr4.tree.Tree import TerminalNodeImpl

from umlmodel.Class import Class
from umlmodel.Method import Method
from umlmodel.Method import Methods
from umlmodel.Method import SourceCode
from umlmodel.Parameter import Parameter
from umlmodel.ParameterType import ParameterType
from umlmodel.ReturnType import ReturnType
from umlmodel.enumerations.Visibility import Visibility

from umlextensions.input.python.pythonpegparser.PythonParser import PythonParser

from umlextensions.input.python.visitor.BaseVisitor import BaseVisitor
from umlextensions.input.python.visitor.BaseVisitor import NO_CLASS_DEF_CONTEXT

from umlextensions.input.python.visitor.ParserTypes import Associate
from umlextensions.input.python.visitor.ParserTypes import AssociateName
from umlextensions.input.python.visitor.ParserTypes import Associates
from umlextensions.input.python.visitor.ParserTypes import AssociationType
from umlextensions.input.python.visitor.ParserTypes import Associations
from umlextensions.input.python.visitor.ParserTypes import Parents
from umlextensions.input.python.visitor.ParserTypes import PropertyMap
from umlextensions.input.python.visitor.ParserTypes import PropertyName
from umlextensions.input.python.visitor.ParserTypes import PropertyNames
from umlextensions.input.python.visitor.ParserTypes import ModelClassName
from umlextensions.input.python.visitor.ParserTypes import ModelClasses
from umlextensions.input.python.visitor.ParentsDictionaryHandler import ParentsDictionaryHandler

MethodName    = NewType('MethodName', str)

"""
    Find 
        the 'def' 
        skip over the method name until we find a '(' 
        then skip until we find ');'
"""

# noinspection SpellCheckingInspection
MAGIC_DUNDER_METHODS:      List[str] = ['__init__', '__str__', '__repr__', '__new__', '__del__',
                                        '__eq__', '__ne__', '__lt__', '__gt__', '__le__', '__ge__'
                                        '__pos__', '__neg__', '__abs__', '__invert__', '__round__', '__floor__', '__floor__', '__trunc__',
                                        '__add__', '__sub__', '__mul__', '__floordiv__', '__div__', '__truediv__', '__mod__', '__divmod__', '__pow__',
                                        '__lshift__', '__rshift__', '__and__', '__or__', '__xor__',
                                        '__hash__',
                                        '__getattr__', '__setattr__', '__getattribute__', '__delattr__',
                                        '__len__', '__setitem__', '__delitem__', '__contains__', '__missing__',
                                        '__call__', '__enter__', '__exit__',
                                        '__bool__'
                                        ]
PARAMETER_SELF:      str = 'self'
PROTECTED_INDICATOR: str = '_'
PRIVATE_INDICATOR:   str = '__'
PROPERTY_DECORATOR:  str = 'property'

"""
    Find 
        the 'def' 
        skip over the method name until we find a '(' 
        then skip until we find ');'
"""
# METHOD_FIND_PATTERN: str = 'def(.*\(.*\):)'     # noqa
METHOD_FIND_PATTERN: str = 'def(.*\x28.*\x29:)'     # noqa


@dataclass
class ParameterNameAndType:
    name:     str = ''
    typeName: str = ''


class PythonPegParserVisitor(BaseVisitor):
    """
    The general purpose Python Code Parser
    """
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

        self._associations: Associations = Associations({})
        self._propertyMap:  PropertyMap  = PropertyMap({})

        self._currentCode:              SourceCode               = cast(SourceCode, None)
        self._parentsDictionaryHandler: ParentsDictionaryHandler = ParentsDictionaryHandler()

    @property
    def modelClasses(self) -> ModelClasses:
        return self._modelClasses

    @modelClasses.setter
    def modelClasses(self, modelClasses: ModelClasses):

        classNames = modelClasses.keys()
        #
        # Create property map for each class
        #
        for className in classNames:
            self._propertyMap[className] = PropertyNames([])
        self._modelClasses = modelClasses

    @property
    def parents(self) -> Parents:
        return self._parentsDictionaryHandler.parents

    @parents.setter
    def parents(self, parents: Parents):
        self._parentsDictionaryHandler.parents = parents

    @property
    def associations(self) -> Associations:
        return self._associations

    @associations.setter
    def associations(self, newValue: Associations):
        self._associations = newValue

    def visitClass_def(self, ctx: PythonParser.Class_defContext):
        """
        Visit a parse tree produced by PythonParser#class_def.

        Args:
            ctx:

        """
        className: ModelClassName = self._extractClassName(ctx=ctx)

        self.logger.debug(f'{className=}')

        argumentsCtx: PythonParser.ArgumentsContext = self._findArgListContext(ctx)
        if argumentsCtx is not None:
            self._parentsDictionaryHandler.createParentChildEntry(argumentsCtx, className)

        return self.visitChildren(ctx)

    def visitFunction_def(self, ctx: PythonParser.Function_defContext):
        """
        Visit a parse tree produced by PythonParser#function_def.

        Args:
            ctx:
        """
        classCtx: PythonParser.Class_defContext = self._extractClassDefContext(ctx)

        if classCtx != NO_CLASS_DEF_CONTEXT:

            className:     ModelClassName = self._extractClassName(ctx=classCtx)
            methodName:    MethodName    = self._extractMethodName(ctx=ctx.function_def_raw())
            returnTypeStr: str           = self._extractReturnType(ctx=ctx)

            visibility: Visibility = Visibility.PUBLIC
            if methodName in MAGIC_DUNDER_METHODS:
                pass
            elif methodName.startswith(PRIVATE_INDICATOR):
                visibility = Visibility.PRIVATE
            elif methodName.startswith(PROTECTED_INDICATOR):
                visibility = Visibility.PROTECTED

            if self._isProperty(ctx) is True:
                self._makePropertyEntry(className=className, methodName=methodName)
                self._handleField(ctx=ctx)
            else:
                self.logger.debug(f'{methodName=}')
                if className not in self._modelClasses:
                    assert False, f'This should not happen missing class name for: {methodName}'
                else:
                    modelClass: Class  = self._modelClasses[className]
                    method:     Method = Method(name=methodName, returnType=ReturnType(returnTypeStr), visibility=visibility)

                    try:
                        method.sourceCode = self._currentCode
                        modelClass.methods.append(method)
                    except Exception as e:
                        self.logger.error(f'{e=}')
                        self.logger.error(f'Missing source code for {className}.{methodName}')

        return self.visitChildren(ctx)

    def visitParameters(self, ctx: PythonParser.ParametersContext):
        """
        Visit a parse tree produced by PythonParser#parameters.

        parameters
            : slash_no_default param_no_default* param_with_default* star_etc?
            | slash_with_default param_with_default* star_etc?
            | param_no_default+ param_with_default* star_etc?
            | param_with_default+ star_etc?
            | star_etc;

        Args:
            ctx:

        Returns:
        """
        classCtx:  PythonParser.Class_defContext    = self._extractClassDefContext(ctx)
        if classCtx == NO_CLASS_DEF_CONTEXT:
            self.logger.warning('This set of parameters belong to a method outside of a class')
        else:
            methodCtx: PythonParser.Function_defContext = self._extractMethodContext(ctx)

            className:    ModelClassName    = self._extractClassName(ctx=classCtx)
            propertyName: PropertyName = self._extractPropertyName(ctx=methodCtx.function_def_raw())
            if self._isThisAParameterListForAProperty(className=className, propertyName=propertyName) is True:
                pass
            else:
                methodName: MethodName = self._extractMethodName(ctx=methodCtx.function_def_raw())
                self.logger.debug(f'{className=} {methodName=}')
                noDefaultContexts: List[PythonParser.Param_no_defaultContext]   = ctx.param_no_default()
                defaultContexts:   List[PythonParser.Param_with_defaultContext] = ctx.param_with_default()

                ctx2 = ctx.slash_no_default()
                ctx3 = ctx.slash_with_default()

                if len(defaultContexts) != 0:
                    self._handleFullParameters(className=className, methodName=methodName, defaultContexts=defaultContexts)
                elif len(noDefaultContexts) != 0:
                    self._handleTypeAnnotated(className=className, methodName=methodName, noDefaultContexts=noDefaultContexts)
                elif ctx2 is not None:
                    self.logger.error(f'{ctx2.getText()}')
                    assert False, f'Unhandled {ctx2.getText()}'
                elif ctx3 is not None:
                    self.logger.error(f'{ctx3.getText()}')
                    assert False, f'Unhandled {ctx3.getText()}'

        return self.visitChildren(ctx)

    def visitAssignment(self, ctx: PythonParser.AssignmentContext):
        """
        Visit a parse tree produced by PythonParser#assignment.
        Do data classes

        Args:
            ctx:
        """
        classDefContext: PythonParser.Class_defContext = self._extractClassDefContext(ctx=ctx)

        if classDefContext != NO_CLASS_DEF_CONTEXT:

            if self._isThisAnAssignmentForADataClass(ctx=classDefContext) is True and self._isThisAssignmentInsideAMethod(ctx=ctx) is False:

                className: ModelClassName = self._extractClassName(ctx=classDefContext)
                self.logger.debug(f'{className} is a data class')
                if len(ctx.children) == 5:
                    self._handleFullField(className, ctx)

                elif len(ctx.children) == 3:
                    # if isinstance(ctx.children[0], TerminalNodeImpl):
                    if isinstance(ctx.children[0], PythonParser.NameContext):
                        self._handleNoDefaultValueField(className, ctx)
                    else:
                        self._handleNoTypeSpecifiedField(className, ctx)

        return self.visitChildren(ctx)

    def visitStatements(self, ctx: PythonParser.StatementsContext):

        parentCtx = ctx.parentCtx

        if isinstance(parentCtx, PythonParser.BlockContext):
            blockContext: PythonParser.BlockContext      = cast(PythonParser.BlockContext, parentCtx)
            statements:   PythonParser.StatementsContext = blockContext.statements()

            for child in statements.children:

                statement:     PythonParser.StatementContext = cast(PythonParser.StatementContext, child)
                statementText: str = statement.start.getInputStream().getText(statement.start.start, statement.stop.stop)

                match: regExMatch | None = regExSearch(METHOD_FIND_PATTERN, statementText)

                if match is not None:

                    self.logger.debug(f'statementText:\n{statementText}')

                    splitStatements: List[str] = statementText.split('\n')

                    sourceCode: SourceCode = SourceCode([])
                    for s in splitStatements:
                        s = s.removeprefix('    ')
                        sourceCode.append(f'{s}')
                    self._currentCode = sourceCode

        return self.visitChildren(ctx)

    def _extractMethodName(self, ctx: PythonParser.Function_def_rawContext) -> MethodName:

        methodName: MethodName       = MethodName(self._extractFunctionNameRawString(ctx=ctx))
        return methodName

    def _extractFunctionNameRawString(self, ctx: PythonParser.Function_def_rawContext) -> str:

        name: TerminalNodeImpl = ctx.name()
        return name.getText()

    def _isProperty(self, ctx: PythonParser.Function_defContext) -> bool:
        """
        Used by the function definition visitor to determine if the current method name is marked as a property.

        Args:
            ctx:  The function's raw context

        Returns: 'True' if it is an annotated property, else 'False'
        """
        ans: bool = False

        decorators: PythonParser.DecoratorsContext = ctx.decorators()
        if decorators is None:
            pass
        else:
            namedExpressions: List[PythonParser.Named_expressionContext] = decorators.named_expression()
            for ne in namedExpressions:
                self.logger.debug(f'{ne.getText()=}')
                if ne.getText() == PROPERTY_DECORATOR:
                    ans = True
                    break
        return ans

    def _extractReturnType(self, ctx: PythonParser.Function_defContext) -> str:

        exprCtx: PythonParser.ExpressionContext = ctx.function_def_raw().expression()

        if exprCtx is None:
            returnTypeStr: str = ''
        else:
            returnTypeStr = exprCtx.getText()

        return returnTypeStr

    def _makePropertyEntry(self, className: ModelClassName, methodName: MethodName):
        """
        Make an entry into the property map.  This ensures that we do not try to create
        arguments for an annotated method when we visit the method parameters

        Args:
            methodName:  A property name which we turn into a field

        """
        self._propertyMap[className].append(cast(PropertyName, methodName))

    def _handleField(self, ctx: PythonParser.Function_defContext):
        """
        Turns methods annotated as a property into an UML field
        Also check to see if it needs to make a entry into the association dictionary

        Args:
            ctx:
        """

        classCtx:  PythonParser.Class_defContext    = self._extractClassDefContext(ctx)
        methodCtx: PythonParser.Function_defContext = self._extractMethodContext(ctx)

        className:    ModelClassName  = self._extractClassName(ctx=classCtx)
        propertyName: PropertyName   = self._extractPropertyName(ctx=methodCtx.function_def_raw())
        self.logger.debug(f'{className} property name: {propertyName}')
        #
        # it is really a property name
        #
        typeStr: str = self._extractReturnType(ctx=ctx)

        self._makeFieldForClass(className, propertyName, typeStr, defaultValue='')

        self._makeAssociationEntry(className, typeStr)

    def _extractPropertyName(self, ctx: PythonParser.Function_def_rawContext) -> PropertyName:

        propertyName: PropertyName = PropertyName(self._extractFunctionNameRawString(ctx=ctx))
        return propertyName

    def _makeAssociationEntry(self, className, typeStr):
        """
        Now check to see if this type is one of our known classes;  If so, then create
        an association entry

        Args:
            className:
            typeStr:

        """
        if typeStr in self._modelClasses:

            associateName: AssociateName = AssociateName(typeStr)
            associate:     Associate     = Associate(associateName=associateName, associationType=AssociationType.ASSOCIATION)

            if className in self._associations:
                self._associations[className].append(associate)
            else:
                self._associations[className] = Associates([associate])

    def _isThisAnAssignmentForADataClass(self, ctx: PythonParser.Class_defContext) -> bool:

        ans: bool = False

        decoratorsCtx: PythonParser.DecoratorsContext = ctx.decorators()
        if decoratorsCtx is not None:
            for decorator in decoratorsCtx.children:
                if isinstance(decorator, PythonParser.Named_expressionContext) is True:
                    # self.logger.info(f'{decorator.getText()=}')
                    ans = True
                    break
        return ans

    def _handleFullField(self, className: ModelClassName, ctx: PythonParser.AssignmentContext):
        """
        From within a data class
        Full annotated and with default value
        Args:
            className:
            ctx:
        """
        fieldName:  str = ctx.children[0].getText()
        typeStr:    str = ctx.children[2].getText()
        fieldValue: str = ctx.children[4].getText()

        self._makeFieldForClass(className=className, propertyName=fieldName, typeStr=typeStr, defaultValue=fieldValue)
        self._makeAssociationEntry(className=className, typeStr=typeStr)

    def _handleNoDefaultValueField(self, className: ModelClassName, ctx: PythonParser.AssignmentContext):
        """
        From inside a data class
        no default value

        Args:
            className:
            ctx:
        """
        fieldName: str = ctx.children[0].getText()
        typeStr:   str = ctx.children[2].getText()

        self._makeFieldForClass(className=className, propertyName=fieldName, typeStr=typeStr, defaultValue='')
        self._makeAssociationEntry(className=className, typeStr=typeStr)

    def _handleNoTypeSpecifiedField(self, className: ModelClassName, ctx: PythonParser.AssignmentContext):
        """
        From inside a data class

        Args:
            className:
            ctx:
        """
        fieldName:    str = ctx.children[0].getText()
        defaultValue: str = ctx.children[2].getText()

        self._makeFieldForClass(className=className, propertyName=fieldName, typeStr='', defaultValue=defaultValue)

    def _isThisAParameterListForAProperty(self, className: ModelClassName, propertyName: PropertyName):
        ans: bool = False

        propertyNames: PropertyNames = self._propertyMap[className]
        if propertyName in propertyNames:
            ans = True

        return ans

    def _handleFullParameters(self, className: ModelClassName, methodName: MethodName, defaultContexts: List[PythonParser.Param_with_defaultContext]):
        """
        Handles these type:
            fullScale(self, intParameter: int = 0, floatParameter: float = 42.0, stringParameter: str = ''):
        """

        for withDefaultCtx in defaultContexts:

            paramCtx:          PythonParser.ParamContext              = withDefaultCtx.param()
            nameAndType:       ParameterNameAndType                   = self._extractParameterNameAndType(paramCtx=paramCtx)
            defaultAssignment: PythonParser.Default_assignmentContext = withDefaultCtx.default_assignment()
            expr:               str                                   = defaultAssignment.children[1].getText()

            parameter: Parameter = Parameter(name=nameAndType.name, type=ParameterType(nameAndType.typeName), defaultValue=expr)
            self._updateModelMethodParameter(className=className, methodName=methodName, parameter=parameter)

    def _handleTypeAnnotated(self, className: ModelClassName, methodName: MethodName, noDefaultContexts: List[PythonParser.Param_no_defaultContext]):

        for noDefaultCtx in noDefaultContexts:
            paramCtx:    PythonParser.ParamContext = noDefaultCtx.param()
            nameAndType: ParameterNameAndType      = self._extractParameterNameAndType(paramCtx=paramCtx)

            if nameAndType.name == PARAMETER_SELF:
                continue

            parameter: Parameter = Parameter(name=nameAndType.name, type=ParameterType(nameAndType.typeName))

            self._updateModelMethodParameter(className=className, methodName=methodName, parameter=parameter)

    def _extractParameterNameAndType(self, paramCtx: PythonParser.ParamContext) -> ParameterNameAndType:

        terminalNode:  TerminalNodeImpl = paramCtx.children[0]
        if len(paramCtx.children) > 1:
            annotationCtx: PythonParser.AnnotationContext = paramCtx.children[1]
            exprCtx:       PythonParser.ExpressionContext = annotationCtx.children[1]
            typeStr: str = exprCtx.getText()
        else:
            typeStr = ''

        paramName: str = terminalNode.getText()

        return ParameterNameAndType(name=paramName, typeName=typeStr)

    def _updateModelMethodParameter(self, className: ModelClassName, methodName: MethodName, parameter: Parameter):

        self.logger.debug(f'{parameter=}')

        modelClass: Class  = self._modelClasses[className]
        method:     Method = self._findModelMethod(methodName=methodName, modelClass=modelClass)

        method.addParameter(parameter=parameter)

    def _findModelMethod(self, modelClass: Class, methodName: MethodName) -> Method:

        foundMethod: Method = cast(Method, None)

        methods: Methods = modelClass.methods
        for m in methods:
            method: Method = cast(Method, m)
            if method.name == methodName:
                foundMethod = method
                break

        return foundMethod




from typing import NewType

ExtensionName        = NewType('ExtensionName', str)
FormatName           = NewType('FormatName', str)
FileSuffix           = NewType('FileSuffix', str)
ExtensionDescription = NewType('ExtensionDescription', str)

Author  = NewType('Author', str)
Version = NewType('Version', str)




from abc import ABC

from umlextensions.extensiontypes.ExtensionDataTypes import ExtensionDescription
from umlextensions.extensiontypes.ExtensionDataTypes import FileSuffix
from umlextensions.extensiontypes.ExtensionDataTypes import FormatName

from umlextensions.extensiontypes.InvalidPluginExtensionException import InvalidPluginExtensionException
from umlextensions.extensiontypes.InvalidPluginNameException import InvalidPluginNameException

DOT:                str = '.'
SPECIAL_CHARACTERS: str = '!@#$%^&*_+-=[]{};:,.<>?/|\'\"'


class BaseFormat(ABC):
    """
    Provides the basic capabilities;  Should not be directly instantiated;
    TODO:  Figure out how to prevent that
    https://stackoverflow.com/questions/7989042/preventing-a-class-from-direct-instantiation-in-python#7990308
    If we do the above; have to figure out how to tests the base functionality in TestBaseFormat
    """
    def __init__(self, formatName: FormatName, fileSuffix: FileSuffix, description: ExtensionDescription):

        if self.__containsSpecialCharacters(formatName):  # TODO Must be a better way
            raise InvalidPluginNameException(f'{formatName}')

        if DOT in fileSuffix:
            raise InvalidPluginExtensionException(f'{fileSuffix}')

        self._name:        FormatName           = formatName
        self._fileSuffix:  FileSuffix           = fileSuffix
        self._description: ExtensionDescription = description

    @property
    def formatName(self) -> FormatName:
        """
        No special characters allowed

        Returns: The Extension's name
        """
        return self._name

    @property
    def fileSuffix(self) -> FileSuffix:
        """
        Returns: The file name suffix (w/o the leading dot '.')
        """
        return self._fileSuffix

    @property
    def description(self) -> ExtensionDescription:
        """
        Returns: The textual description of the extension data  format
        """
        return self._description

    def __containsSpecialCharacters(self, name: FormatName) -> bool:
        for special in SPECIAL_CHARACTERS:
            if special in name:
                return True
        return False




from typing import cast
from typing import List

from logging import Logger
from logging import getLogger

from os import linesep as osLineSep

from antlr4 import ParserRuleContext
from umlmodel.Class import Class
from umlmodel.enumerations.Stereotype import Stereotype

from umlextensions.input.python.pythonpegparser.PythonParser import PythonParser

from umlextensions.input.python.visitor.BaseVisitor import BaseVisitor
from umlextensions.input.python.visitor.BaseVisitor import NO_CLASS_DEF_CONTEXT
from umlextensions.input.python.visitor.ParserTypes import ParentName
from umlextensions.input.python.visitor.ParserTypes import ModelClassName

from umlextensions.input.python.visitor.ParserTypes import ModelClasses
from umlextensions.input.python.visitor.ParserTypes import VERSION

ENUMERATION_SUPER_CLASS: str = 'Enum'


class PythonPegParserClassVisitor(BaseVisitor):
    """
    The UML Extension visitor specific to the Python Input Extension

    Simply does a scan to identify all the classes;   A separate
    is needed to do inheritance

    """
    def __init__(self):
        super().__init__()
        self.logger: Logger = getLogger(__name__)

    @property
    def modelClasses(self) -> ModelClasses:
        return self._modelClasses

    @modelClasses.setter
    def modelClasses(self, modelClasses: ModelClasses):
        self._modelClasses = modelClasses

    def visitClass_def(self, ctx: PythonParser.Class_defContext):
        """
        Visit a parse tree produced by PythonParser#class_def.

        Args:
            ctx:
        """
        #
        # Check if we are an enumeration
        #
        className: ModelClassName = self._extractClassName(ctx=ctx)

        modelClass: Class = Class(name=className)
        modelClass.description = self._generateMyCredits()

        argumentsCtx: PythonParser.ArgumentsContext = self._findArgListContext(ctx)

        if argumentsCtx is not None:
            args: PythonParser.ArgsContext = argumentsCtx.args()
            parentName: ParentName = ParentName(args.getText())
            self.logger.debug(f'Class: {className} is subclass of {parentName}')
            parents: List[str] = parentName.split(',')
            for parent in parents:
                if parent == ENUMERATION_SUPER_CLASS:
                    modelClass.stereotype = Stereotype.ENUMERATION
                    break

        self._modelClasses[className] = modelClass

        return self.visitChildren(ctx)

    def visitAssignment(self, ctx: PythonParser.AssignmentContext):
        """
        Visit a parse tree produced by PythonParser#assignment.

        Args:
            ctx:
        """
        if self._isThisAssignmentInsideAMethod(ctx=ctx) is False:

            classCtx:  PythonParser.Class_defContext = self._extractClassDefContext(ctx)
            if classCtx == NO_CLASS_DEF_CONTEXT:
                pass
            else:
                className: ModelClassName = self._extractClassName(ctx=classCtx)
                modelClass: Class         = self._modelClasses[className]

                if modelClass.stereotype == Stereotype.ENUMERATION:
                    if len(ctx.children) >= 2:
                        enumName:     str = ctx.children[0].getText()
                        defaultValue: str = ctx.children[2].getText()
                        self.logger.info(f'')
                        self._makeFieldForClass(className=className, propertyName=enumName, typeStr='', defaultValue=defaultValue)

        return self.visitChildren(ctx)

    def visitPrimary(self, ctx: PythonParser.PrimaryContext):
        """
        Generates artificial/synthetic types
        Args:
            ctx:
        """
        primaryStr: str = ctx.getText()
        if primaryStr.startswith('NewType'):
            argumentsCtx: PythonParser.ArgumentsContext = ctx.arguments()
            if argumentsCtx is not None:

                argStr = ctx.children[2].getText()
                typeValueList = argStr.split(',')
                self.logger.debug(f'{typeValueList=}')

                className = typeValueList[0].strip("'").strip('"')
                self.logger.debug(f'{className}')

                modelClass: Class = Class(name=className)

                modelClass.description = self._generateMyCredits()
                modelClass.stereotype  = Stereotype.TYPE

                self._modelClasses[className] = modelClass

        return self.visitChildren(ctx)

    def _isThisAssignmentInsideAMethod(self, ctx: PythonParser.AssignmentContext) -> bool:

        ans: bool = False

        currentCtx: ParserRuleContext = self._extractMethodContext(ctx=ctx)
        if currentCtx is not None:
            ans = True

        return ans

    def _extractMethodContext(self, ctx: ParserRuleContext) -> PythonParser.Function_defContext:

        currentCtx: ParserRuleContext = ctx

        while isinstance(currentCtx, PythonParser.Function_defContext) is False:
            currentCtx = currentCtx.parentCtx
            if currentCtx is None:
                break

        if currentCtx is not None:
            raw: PythonParser.Function_def_rawContext = cast(PythonParser.Function_defContext, currentCtx).function_def_raw()
            # self.baseLogger.debug(f'Found method: {raw.NAME()}')
            self.baseLogger.debug(f'Found method: {raw.name()}')

        return cast(PythonParser.Function_defContext, currentCtx)

    def _extractClassDefContext(self, ctx: ParserRuleContext) -> PythonParser.Class_defContext:
        """
        Args:
            ctx:

        Returns:  Either a class definition context or the sentinel value NO_CLASS_DEF_CONTEXT
        """
        currentCtx: ParserRuleContext = ctx
        while currentCtx.parentCtx:
            if isinstance(currentCtx, PythonParser.Class_defContext):
                return currentCtx
            currentCtx = currentCtx.parentCtx

        return NO_CLASS_DEF_CONTEXT

    def _generateMyCredits(self) -> str:
        """

        Returns:    Reversed Engineered by the one and only:
                    Gato Malo - Humberto A. Sanchez II
                    Generated: ${DAY} ${MONTH_NAME_FULL} ${YEAR}
                    Version: ${VERSION}

        """
        from datetime import date

        today: date = date.today()
        formatDated: str = today.strftime('%d %B %Y')

        hasiiCredits: str = (
            f'Reversed Engineered by the one and only:{osLineSep}'
            f'Gato Malo - Humberto A. Sanchez II{osLineSep}'
            f'Generated: {formatDated}{osLineSep}'
            f'Version: {VERSION}'
        )

        return hasiiCredits



#
#  These types are visible to the extension adapter and the
#  implementor of the adapter
#
from typing import cast
from typing import Callable

from dataclasses import dataclass
from dataclasses import field

from wx import ClientDC

from umlshapes.frames.UmlFrame import UmlFrame

from umlshapes.ShapeTypes import UmlShapes
from umlshapes.ShapeTypes import umlShapesFactory


@dataclass
class FrameSize:
    """
    The strategy is to provide minimal information to the extensions
    we do not want them to not abuse it.
    """
    width:  int = -1
    height: int = -1


def createFrameSizeFactory() -> FrameSize:
    """
    Factory method to create  the OglClasses data structure;

    Returns:  A new data structure
    """
    return FrameSize()


@dataclass
class FrameInformation:
    """
    The document title is the name of the frame
    """
    umlFrame:           UmlFrame   = cast(UmlFrame, None)
    frameActive:        bool       = False
    selectedUmlShapes:  UmlShapes  = field(default_factory=umlShapesFactory)
    diagramTitle:       str        = ''
    diagramType:        str        = ''
    frameSize:          FrameSize  = field(default_factory=createFrameSizeFactory)
    clientDC:           ClientDC   = cast(ClientDC, None)


FrameInformationCallback  = Callable[[FrameInformation], None]
FrameSizeCallback         = Callable[[FrameSize], None]
SelectedUmlShapesCallback = Callable[[UmlShapes], None]

NO_INTEGER: int = cast(int, None)




from logging import Logger
from logging import getLogger

from codeallybasic.SecureConversions import SecureConversions

from codeallybasic.SingletonV3 import SingletonV3

from codeallybasic.DynamicConfiguration import Sections
from codeallybasic.DynamicConfiguration import KeyName
from codeallybasic.DynamicConfiguration import SectionName
from codeallybasic.DynamicConfiguration import ValueDescription
from codeallybasic.DynamicConfiguration import ValueDescriptions
from codeallybasic.DynamicConfiguration import DynamicConfiguration

MODULE_NAME:          str = 'umlextensions'
PREFERENCES_FILENAME: str = f'{MODULE_NAME}.ini'

SECTION_EXTENSIONS: ValueDescriptions = ValueDescriptions(
    {
        KeyName('sugiyamaStepByStep'): ValueDescription(defaultValue='False', deserializer=SecureConversions.secureBoolean),
    }
)

SECTION_FEATURES: ValueDescriptions = ValueDescriptions(
    {
        KeyName('startDirectory'):           ValueDescription(defaultValue=''),
    }
)

SECTION_SHAPE_LAYOUT: ValueDescriptions = ValueDescriptions(
    {
        KeyName('startX'):     ValueDescription(defaultValue='20',   deserializer=SecureConversions.secureInteger),
        KeyName('startY'):     ValueDescription(defaultValue='20',   deserializer=SecureConversions.secureInteger),
        KeyName('xIncrement'): ValueDescription(defaultValue='20',   deserializer=SecureConversions.secureInteger),
        KeyName('maximumX'):   ValueDescription(defaultValue='3000', deserializer=SecureConversions.secureInteger),
    }
)

SECTION_DEBUG: ValueDescriptions = ValueDescriptions(
    {
        KeyName('autoSelectAll'): ValueDescription(defaultValue='True', deserializer=SecureConversions.secureBoolean),
    }
)

EXTENSION_SECTIONS: Sections = Sections(
    {
        SectionName('Extensions'):   SECTION_EXTENSIONS,
        SectionName('Features'):     SECTION_FEATURES,
        SectionName('Shape Layout'): SECTION_SHAPE_LAYOUT,
        SectionName('Debug'):        SECTION_DEBUG,
    }
)


class ExtensionsPreferences(DynamicConfiguration, metaclass=SingletonV3):
    def __init__(self):
        self._logger: Logger = getLogger(__name__)

        super().__init__(baseFileName=f'{PREFERENCES_FILENAME}', moduleName=MODULE_NAME, sections=EXTENSION_SECTIONS)

        self._configParser.optionxform = str  # type: ignore



# The MIT License (MIT)
# Copyright (c) 2021 Robert Einhorn
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Project      : Python Indent/Dedent handler for ANTLR4 grammars
# 
# Developed by : Robert Einhorn

from typing import TextIO, Optional, List, Deque
from antlr4 import InputStream, Lexer, Token
from antlr4.Token import CommonToken
import sys
import re

class PythonLexerBase(Lexer):
    def __init__(self, input: InputStream, output: TextIO = sys.stdout):
        super().__init__(input, output)

        # A stack that keeps track of the indentation lengths
        self.__indent_length_stack: List[int]

        # A list where tokens are waiting to be loaded into the token stream
        self.__pending_tokens: Deque[CommonToken]

        # last pending token type
        self.__previous_pending_token_type: int
        self.__last_pending_token_type_from_default_channel: int

        # The amount of opened parentheses, square brackets or curly braces
        self.__opened: int
        # The amount of opened parentheses and square brackets in the current lexer mode
        self.__paren_or_bracket_opened_stack: List[int]
        # A stack that stores expression(s) between braces in fstring
        self.__brace_expression_stack: List[str]
        self.__prev_brace_expression: str
        
        # Instead of self._mode      (self._mode is not implemented in each ANTLR4 runtime)
        self.__cur_lexer_mode: int
        # Instead of self._modeStack (self._modeStack is not implemented in each ANTLR4 runtime)
        self.__lexer_mode_stack: List[int]

        self.__was_space_indentation: bool
        self.__was_tab_indentation: bool
        self.__was_indentation_mixed_with_spaces_and_tabs: bool

        self.__cur_token: CommonToken # current (under processing) token
        self.__ffg_token: CommonToken # following (look ahead) token

        self.__INVALID_LENGTH: int = -1
        self.__ERR_TXT: str = " ERROR: "

        self.__init()

    def nextToken(self) -> CommonToken: # reading the input stream until a return EOF
        self.__check_next_token()
        return self.__pending_tokens.popleft() # add the queued token to the token stream

    def reset(self) -> None:
        self.__init()
        super().reset()

    def __init(self) -> None:
        self.__indent_length_stack = []
        self.__pending_tokens = Deque()
        self.__previous_pending_token_type = 0
        self.__last_pending_token_type_from_default_channel = 0
        self.__opened = 0
        self.__paren_or_bracket_opened_stack = []
        self.__brace_expression_stack = []
        self.__prev_brace_expression = ""
        self.__cur_lexer_mode = 0
        self.__lexer_mode_stack = []
        self.__was_space_indentation = False
        self.__was_tab_indentation = False
        self.__was_indentation_mixed_with_spaces_and_tabs = False
        self.__cur_token = None
        self.__ffg_token = None

    def __check_next_token(self) -> None:
        if self.__previous_pending_token_type == Token.EOF:
            return

        if not self.__indent_length_stack: # We're at the first token
            self.__insert_ENCODING_token()
            self.__set_current_and_following_tokens()
            self.__handle_start_of_input()
        else:
            self.__set_current_and_following_tokens()

        match self.__cur_token.type:
            case self.NEWLINE:
                self.__handle_NEWLINE_token()                
            case self.LPAR | self.LSQB | self.LBRACE:
                self.__opened += 1
                self.__add_pending_token(self.__cur_token)
            case self.RPAR | self.RSQB | self.RBRACE:
                self.__opened -= 1
                self.__add_pending_token(self.__cur_token)
            case self.FSTRING_MIDDLE:
                self.__handle_FSTRING_MIDDLE_token_with_double_brace() # does not affect the opened field
                self.__add_pending_token(self.__cur_token)
            case self.COLONEQUAL:
                self.__handle_COLONEQUAL_token_in_fstring()
            case self.ERRORTOKEN:
                self.__report_lexer_error("token recognition error at: '" + self.__cur_token.text + "'")
                self.__add_pending_token(self.__cur_token)
            case Token.EOF:
                self.__handle_EOF_token()
            case _:
                self.__add_pending_token(self.__cur_token)
        self.__handle_FORMAT_SPECIFICATION_MODE()

    def __set_current_and_following_tokens(self) -> None:
        self.__cur_token = super().nextToken() if self.__ffg_token is None else \
                           self.__ffg_token

        self.__check_cur_token() # ffgToken cannot be used in this method and its sub methods (ffgToken is not yet set)!
        
        self.__ffg_token = self.__cur_token if self.__cur_token.type == Token.EOF else \
                           super().nextToken()

    def __insert_ENCODING_token(self) -> None:  # https://peps.python.org/pep-0263/
        line_builder: list[str] = []
        encoding_name: str = ""
        line_count: int = 0
        ws_comment_pattern: re.Pattern = re.compile(r"^[ \t\f]*(#.*)?$")
        input_stream: InputStream = self.inputStream
        size: int = input_stream.size

        input_stream.seek(0)
        for i in range(size):
            c: str = chr(input_stream.LA(i + 1))
            line_builder.append(c)

            if c == '\n' or i == size - 1:
                line: str = ''.join(line_builder).replace("\r", "").replace("\n", "")
                if ws_comment_pattern.match(line): # WS* + COMMENT? found
                    encoding_name = self.__get_encoding_name(line)
                    if encoding_name:
                        break # encoding found
                else:
                    break # statement or backslash found (first line is not empty, not whitespace(s), not comment)
                
                line_count += 1
                if line_count >= 2:
                    break # check only the first two lines
                line_builder = []

        if not encoding_name:
            encoding_name = "utf-8"  # default Python source code encoding

        encoding_token: CommonToken = CommonToken((None, None), self.ENCODING, CommonToken.HIDDEN_CHANNEL, 0, 0)
        encoding_token.text = encoding_name
        encoding_token.line = 0
        encoding_token.column = -1
        self.__add_pending_token(encoding_token)

    def __get_encoding_name(self, comment_text: str) -> str:  # https://peps.python.org/pep-0263/#defining-the-encoding
        encoding_comment_pattern: str = r"^[ \t\f]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+)"
        match: Optional[re.Match] = re.search(encoding_comment_pattern, comment_text)
        return match.group(1) if match else ""

    # initialize the _indent_length_stack
    # hide the leading NEWLINE token(s)
    # if exists, find the first statement (not NEWLINE, not EOF token) that comes from the default channel
    # insert a leading INDENT token if necessary
    def __handle_start_of_input(self) -> None:
        # initialize the stack with a default 0 indentation length
        self.__indent_length_stack.append(0) # this will never be popped off
        while self.__cur_token.type != Token.EOF:
            if self.__cur_token.channel == Token.DEFAULT_CHANNEL:
                if self.__cur_token.type == self.NEWLINE:
                    # all the NEWLINE tokens must be ignored before the first statement
                    self.__hide_and_add_pending_token(self.__cur_token)
                else: # We're at the first statement
                    self.__insert_leading_indent_token()
                    return # continue the processing of the current token with __check_next_token()
            else:
                self.__add_pending_token(self.__cur_token) # it can be WS, EXPLICIT_LINE_JOINING or COMMENT token
            self.__set_current_and_following_tokens()
        # continue the processing of the EOF token with __check_next_token()

    def __insert_leading_indent_token(self) -> None:
        if self.__previous_pending_token_type == self.WS:
            prev_token: CommonToken = self.__pending_tokens[-1]  # WS token
            if self.__get_indentation_length(prev_token.text) != 0: # there is an "indentation" before the first statement
                err_msg: str = "first statement indented"
                self.__report_lexer_error(err_msg)
                # insert an INDENT token before the first statement to raise an 'unexpected indent' error later by the parser
                self.__create_and_add_pending_token(self.INDENT, Token.DEFAULT_CHANNEL, self.__ERR_TXT + err_msg, self.__cur_token)

    def __handle_NEWLINE_token(self) -> None:
        if self.__lexer_mode_stack: # not is_empty
            self.__add_pending_token(self.__cur_token)
        elif self.__opened > 0: # We're in an implicit line joining, ignore the current NEWLINE token
            self.__hide_and_add_pending_token(self.__cur_token)
        else:
            nl_token: CommonToken = self.__cur_token.clone() # save the current NEWLINE token
            is_looking_ahead: bool = self.__ffg_token.type == self.WS
            if is_looking_ahead:
                self.__set_current_and_following_tokens() # set the next two tokens

            match self.__ffg_token.type:
                case self.NEWLINE | self.COMMENT:
                    # We're before a blank line or a comment or type comment or a type ignore comment
                    self.__hide_and_add_pending_token(nl_token) # ignore the NEWLINE token
                    if is_looking_ahead:
                        self.__add_pending_token(self.__cur_token) # WS token
                case _:
                    self.__add_pending_token(nl_token)
                    if is_looking_ahead: # We're on a whitespace(s) followed by a statement
                        indentation_length: int = 0 if self.__ffg_token.type == Token.EOF else \
                                                  self.__get_indentation_length(self.__cur_token.text)

                        if indentation_length != self.__INVALID_LENGTH:
                            self.__add_pending_token(self.__cur_token) # WS token
                            self.__insert_INDENT_or_DEDENT_token(indentation_length) # may insert INDENT token or DEDENT token(s)
                        else:
                            self.__report_error("inconsistent use of tabs and spaces in indentation")
                    else: # We're at a newline followed by a statement (there is no whitespace before the statement)
                        self.__insert_INDENT_or_DEDENT_token(0) # may insert DEDENT token(s)

    def __insert_INDENT_or_DEDENT_token(self, indent_length: int) -> None:
        prev_indent_length: int = self.__indent_length_stack[-1] # stack peek
        if indent_length > prev_indent_length:
            self.__create_and_add_pending_token(self.INDENT, Token.DEFAULT_CHANNEL, None, self.__ffg_token)
            self.__indent_length_stack.append(indent_length) # stack push
        else:
            while indent_length < prev_indent_length: # more than 1 DEDENT token may be inserted to the token stream
                self.__indent_length_stack.pop()
                prev_indent_length = self.__indent_length_stack[-1] # stack peek
                if indent_length <= prev_indent_length:
                    self.__create_and_add_pending_token(self.DEDENT, Token.DEFAULT_CHANNEL, None, self.__ffg_token)
                else:
                    self.__report_error("inconsistent dedent")

    def __check_cur_token(self) -> None:
        match self.__cur_token.type:
            case self.FSTRING_START:
                self.__set_lexer_mode_by_FSTRING_START_token()
                return
            case self.FSTRING_MIDDLE:
                self.__handle_FSTRING_MIDDLE_token_with_quote_and_lbrace() # affect the opened field
                if self.__cur_token.type == self.FSTRING_MIDDLE:
                    return  # No __cur_token exchange happened
            case self.FSTRING_END:
                self.__pop_lexer_mode()
                return
            case _:
                if not self.__lexer_mode_stack:
                    return  # Not in fstring mode

        match self.__cur_token.type:  # the following tokens can only come from default mode (after an LBRACE in fstring)
            case self.NEWLINE:
                # append the current brace expression with the current newline
                self.__append_to_brace_expression(self.__cur_token.text)
                self.__cur_token.channel = Token.HIDDEN_CHANNEL
            case self.LBRACE:
                # the outermost brace expression cannot be a dictionary comprehension or a set comprehension
                self.__brace_expression_stack.append("{")
                self.__paren_or_bracket_opened_stack.append(0) # stack push
                self.__push_lexer_mode(Lexer.DEFAULT_MODE)
            case self.LPAR | self.LSQB:
                # append the current brace expression with a "(" or a "["
                self.__append_to_brace_expression(self.__cur_token.text)
                # https://peps.python.org/pep-0498/#lambdas-inside-expressions
                self.__increment_brace_stack()
            case self.RPAR | self.RSQB:
                # append the current brace expression with a ")" or a "]"
                self.__append_to_brace_expression(self.__cur_token.text)
                self.__decrement_brace_stack()
            case self.COLON | self.COLONEQUAL:
                # append the current brace expression with a ":" or a ":="
                self.__append_to_brace_expression(self.__cur_token.text)
                self.__set_lexer_mode_by_COLON_or_COLONEQUAL_token()
            case self.RBRACE:
                self.__set_lexer_mode_after_RBRACE_token()
            case _:
                # append the current brace expression with the current token text
                self.__append_to_brace_expression(self.__cur_token.text)

    def __append_to_brace_expression(self, text: str) -> None:
        self.__brace_expression_stack[-1] += text

    def __increment_brace_stack(self) -> None: # increment the last element (peek() + 1)
        self.__paren_or_bracket_opened_stack[-1] += 1

    def __decrement_brace_stack(self) -> None: # decrement the last element (peek() - 1)
        self.__paren_or_bracket_opened_stack[-1] -= 1

    def __set_lexer_mode_after_RBRACE_token(self) -> None:
        match self.__cur_lexer_mode:
            case Lexer.DEFAULT_MODE:
                self.__pop_lexer_mode() # only once
                self.__pop_by_RBRACE()
            case self.SQ1__FORMAT_SPECIFICATION_MODE \
                | self.SQ1R_FORMAT_SPECIFICATION_MODE \
                | self.DQ1__FORMAT_SPECIFICATION_MODE \
                | self.DQ1R_FORMAT_SPECIFICATION_MODE \
                | self.SQ3__FORMAT_SPECIFICATION_MODE \
                | self.SQ3R_FORMAT_SPECIFICATION_MODE \
                | self.DQ3__FORMAT_SPECIFICATION_MODE \
                | self.DQ3R_FORMAT_SPECIFICATION_MODE:

                self.__pop_lexer_mode()
                self.__pop_lexer_mode()
                self.__pop_by_RBRACE()
            case _:
                self.__report_lexer_error("f-string: single '}' is not allowed")

    def __set_lexer_mode_by_FSTRING_START_token(self) -> None:
        text = self.__cur_token.text.lower()
        mode_map = {
            "f'": self.SQ1__FSTRING_MODE,
            "rf'": self.SQ1R_FSTRING_MODE,
            "fr'": self.SQ1R_FSTRING_MODE,
            'f"': self.DQ1__FSTRING_MODE,
            'rf"': self.DQ1R_FSTRING_MODE,
            'fr"': self.DQ1R_FSTRING_MODE,
            "f'''": self.SQ3__FSTRING_MODE,
            "rf'''": self.SQ3R_FSTRING_MODE,
            "fr'''": self.SQ3R_FSTRING_MODE,
            'f"""': self.DQ3__FSTRING_MODE,
            'rf"""': self.DQ3R_FSTRING_MODE,
            'fr"""': self.DQ3R_FSTRING_MODE,
        }
        mode = mode_map.get(text)
        if mode is not None:
            self.__push_lexer_mode(mode)

    def __set_lexer_mode_by_COLON_or_COLONEQUAL_token(self) -> None:
        if self.__paren_or_bracket_opened_stack[-1] == 0: # stack peek == 0
            # COLONEQUAL token will be replaced with a COLON token in checkNextToken()
            match self.__lexer_mode_stack[-1]: # check the previous lexer mode (the current is DEFAULT_MODE)
                case self.SQ1__FSTRING_MODE \
                    | self.SQ1__FORMAT_SPECIFICATION_MODE:

                    self.__push_lexer_mode(self.SQ1__FORMAT_SPECIFICATION_MODE) # continue in format spec. mode
                case self.SQ1R_FSTRING_MODE \
                    | self.SQ1R_FORMAT_SPECIFICATION_MODE:

                    self.__push_lexer_mode(self.SQ1R_FORMAT_SPECIFICATION_MODE) # continue in format spec. mode
                case self.DQ1__FSTRING_MODE \
                    | self.DQ1__FORMAT_SPECIFICATION_MODE:

                    self.__push_lexer_mode(self.DQ1__FORMAT_SPECIFICATION_MODE) # continue in format spec. mode
                case self.DQ1R_FSTRING_MODE \
                    | self.DQ1R_FORMAT_SPECIFICATION_MODE:

                    self.__push_lexer_mode(self.DQ1R_FORMAT_SPECIFICATION_MODE) # continue in format spec. mode
                case self.SQ3__FSTRING_MODE \
                    | self.SQ3__FORMAT_SPECIFICATION_MODE:
                    
                    self.__push_lexer_mode(self.SQ3__FORMAT_SPECIFICATION_MODE) # continue in format spec. mode
                case self.SQ3R_FSTRING_MODE \
                    | self.SQ3R_FORMAT_SPECIFICATION_MODE:
                    
                    self.__push_lexer_mode(self.SQ3R_FORMAT_SPECIFICATION_MODE) # continue in format spec. mode
                case self.DQ3__FSTRING_MODE \
                    | self.DQ3__FORMAT_SPECIFICATION_MODE:
                    
                    self.__push_lexer_mode(self.DQ3__FORMAT_SPECIFICATION_MODE) # continue in format spec. mode
                case self.DQ3R_FSTRING_MODE \
                    | self.DQ3R_FORMAT_SPECIFICATION_MODE:
                    
                    self.__push_lexer_mode(self.DQ3R_FORMAT_SPECIFICATION_MODE) # continue in format spec. mode

    def __pop_by_RBRACE(self) -> None:
        self.__paren_or_bracket_opened_stack.pop()
        self.__prev_brace_expression = self.__brace_expression_stack.pop() + "}"
        if self.__brace_expression_stack:
            # append the current brace expression with the previous brace expression
            self.__brace_expression_stack[-1] += self.__prev_brace_expression

    def __handle_FSTRING_MIDDLE_token_with_double_brace(self) -> None:
        # replace the trailing double brace with a single brace and insert a hidden brace token
        match self.__get_last_two_chars_of_the_cur_token_text():
            case "{{":
                self.__trim_last_char_add_pending_token_set_cur_token(self.LBRACE, "{", Token.HIDDEN_CHANNEL)
            case "}}":
                self.__trim_last_char_add_pending_token_set_cur_token(self.RBRACE, "}", Token.HIDDEN_CHANNEL)

    def __handle_FSTRING_MIDDLE_token_with_quote_and_lbrace(self) -> None:
        # replace the trailing     quote + left_brace with a quote     and insert an LBRACE token
        # replace the trailing backslash + left_brace with a backslash and insert an LBRACE token        
        match self.__get_last_two_chars_of_the_cur_token_text():
            case "\"{" | "'{" | "\\{":
                self.__trim_last_char_add_pending_token_set_cur_token(self.LBRACE, "{", Token.DEFAULT_CHANNEL)

    def __get_last_two_chars_of_the_cur_token_text(self) -> str:
        cur_token_text: str = self.__cur_token.text
        return cur_token_text[-2:] if len(cur_token_text) >= 2 else cur_token_text

    def __trim_last_char_add_pending_token_set_cur_token(self, type: int, text: str, channel: int) -> None:
        # trim the last char and add the modified curToken to the __pending_tokens stack
        token_text_without_lbrace: str = self.__cur_token.text[:-1]
        self.__cur_token.text = token_text_without_lbrace
        self.__cur_token.stop -= 1
        self.__add_pending_token(self.__cur_token)

        self.__create_new_cur_token(type, text, channel) # set __cur_token

    def __handle_COLONEQUAL_token_in_fstring(self) -> None:
        if self.__lexer_mode_stack \
           and self.__paren_or_bracket_opened_stack[-1] == 0: # stack peek == 0

            # In fstring a colonequal (walrus operator) can only be used in parentheses
            # Not in parentheses, replace COLONEQUAL token with COLON as format specifier
            # and insert the equal symbol to the following FSTRING_MIDDLE token
            self.__cur_token.type = self.COLON
            self.__cur_token.text = ":"
            self.__cur_token.stop = self.__cur_token.start
            if self.__ffg_token.type == self.FSTRING_MIDDLE:
                self.__ffg_token.text = "=" + self.__ffg_token.text
                self.__ffg_token.start -= 1
                self.__ffg_token.column -= 1
            else:
                self.__add_pending_token(self.__cur_token)
                self.__create_new_current_token(self.FSTRING_MIDDLE, "=", Token.DEFAULT_CHANNEL)
        self.__add_pending_token(self.__cur_token)

    def __create_new_cur_token(self, type: int, text: str, channel: int) -> None:
        ctkn: CommonToken  = self.__cur_token.clone()
        ctkn.type  = type
        ctkn.text = text
        ctkn.channel = channel
        ctkn.column += 1
        ctkn.start += 1
        ctkn.stop = ctkn.start
        self.__cur_token = ctkn

    def __push_lexer_mode(self, mode: int) -> None:
        self.pushMode(mode)
        self.__lexer_mode_stack.append(self.__cur_lexer_mode) # stack push
        self.__cur_lexer_mode = mode

    def __pop_lexer_mode(self) -> None:
        self.popMode()
        self.__cur_lexer_mode = self.__lexer_mode_stack.pop()

    def __handle_FORMAT_SPECIFICATION_MODE(self) -> None:
        if self.__lexer_mode_stack \
           and self.__ffg_token.type == self.RBRACE:

            # insert an empty FSTRING_MIDDLE token instead of the missing format specification
            match self.__cur_token.type:
                case self.COLON:
                    self.__create_and_add_pending_token(self.FSTRING_MIDDLE, Token.DEFAULT_CHANNEL, "", self.__ffg_token)
                case self.RBRACE:
                    # only if the previous brace expression is not a dictionary comprehension or set comprehension
                    if not self.__is_dictionary_comprehension_or_set_comprehension(self.__prev_brace_expression):
                        self.__create_and_add_pending_token(self.FSTRING_MIDDLE, Token.DEFAULT_CHANNEL, "", self.__ffg_token)

    def __is_dictionary_comprehension_or_set_comprehension(self, code: str) -> bool:
        from antlr4 import InputStream, CommonTokenStream
        from PythonLexer import PythonLexer
        from PythonParser import PythonParser

        input_stream: InputStream = InputStream(code)
        lexer: PythonLexer = PythonLexer(input_stream)
        token_stream: CommonTokenStream  = CommonTokenStream(lexer)
        parser: PythonParser = PythonParser(token_stream)

        # Disable error listeners to suppress console output
        lexer.removeErrorListeners()  
        parser.removeErrorListeners()

        parser.dictcomp() # Try parsing as dictionary comprehension
        if parser.getNumberOfSyntaxErrors() == 0:
            return True

        parser = PythonParser(token_stream)
        token_stream.seek(0)
        parser.removeErrorListeners()
        parser.setcomp() # Try parsing as set comprehension
        return parser.getNumberOfSyntaxErrors() == 0

    def __insert_trailing_tokens(self) -> None:
        match self.__last_pending_token_type_from_default_channel:
            case self.NEWLINE | self.DEDENT:
                pass # no trailing NEWLINE token is needed
            case _: # insert an extra trailing NEWLINE token that serves as the end of the last statement
                self.__create_and_add_pending_token(self.NEWLINE, Token.DEFAULT_CHANNEL, None, self.__ffg_token) # _ffg_token is EOF
        self.__insert_INDENT_or_DEDENT_token(0) # Now insert as much trailing DEDENT tokens as needed

    def __handle_EOF_token(self) -> None:
        if self.__last_pending_token_type_from_default_channel > 0:
            # there was statement in the input (leading NEWLINE tokens are hidden)
            self.__insert_trailing_tokens()
        self.__add_pending_token(self.__cur_token)

    def __hide_and_add_pending_token(self, ctkn: CommonToken) -> None:
        ctkn.channel = Token.HIDDEN_CHANNEL
        self.__add_pending_token(ctkn)

    def __create_and_add_pending_token(self, ttype: int, channel: int, text: Optional[str], sample_token: CommonToken) -> None:
        ctkn: CommonToken = sample_token.clone()
        ctkn.type  = ttype
        ctkn.channel = channel
        ctkn.stop = sample_token.start - 1
        ctkn.text = "<" + self.symbolicNames[ttype] + ">" if text is None else \
                    text

        self.__add_pending_token(ctkn)

    def __add_pending_token(self, ctkn: CommonToken) -> None:
        # save the last pending token type because the _pending_tokens list can be empty by the nextToken()
        self.__previous_pending_token_type = ctkn.type
        if ctkn.channel == Token.DEFAULT_CHANNEL:
            self.__last_pending_token_type_from_default_channel = self.__previous_pending_token_type
        self.__pending_tokens.append(ctkn)

    def __get_indentation_length(self, indentText: str) -> int: # the indentText may contain spaces, tabs or form feeds
        TAB_LENGTH: int = 8 # the standard number of spaces to replace a tab with spaces
        length: int = 0
        ch: str
        for ch in indentText:
            match ch:
                case ' ':
                    self.__was_space_indentation = True
                    length += 1
                case '\t':
                    self.__was_tab_indentation = True
                    length += TAB_LENGTH - (length % TAB_LENGTH)
                case '\f': # form feed
                    length = 0

        if self.__was_tab_indentation and self.__was_space_indentation:
            if not self.__was_indentation_mixed_with_spaces_and_tabs:
                self.__was_indentation_mixed_with_spaces_and_tabs = True
                length = self.__INVALID_LENGTH # only for the first inconsistent indent
        return length

    def __report_lexer_error(self, err_msg: str) -> None:
        self.getErrorListenerDispatch().syntaxError(self, self.__cur_token, self.__cur_token.line, self.__cur_token.column, " LEXER" + self.__ERR_TXT + err_msg, None)

    def __report_error(self, err_msg: str) -> None:
        self.__report_lexer_error(err_msg)

        # the ERRORTOKEN will raise an error in the parser
        self.__create_and_add_pending_token(self.ERRORTOKEN, Token.DEFAULT_CHANNEL, self.__ERR_TXT + err_msg, self.__ffg_token)
